// Tab.cc for Fluxbox
// Copyright (c) 2001 Henrik Kinnunen (fluxgen@linuxmail.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#include "Tab.hh"
#include <iostream>
#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif // HAVE_CONFIG_H
#include "i18n.hh"

//const size of tab window
unsigned int Tab::m_size_w = 64;
unsigned int Tab::m_size_h = 16;
const int Tab::m_max_tabs = 512;
bool Tab::m_stoptabs = false;

Tab::Tab(FluxboxWindow *win, Tab *prev, Tab *next) { 
	//set default values
	m_focus = m_moving = false;
	m_move_x = m_move_y = 0;	
	m_prev = prev; m_next = next; 
	m_win = win;
	m_display = Fluxbox::instance()->getXDisplay();
	createTabWindow();
}

Tab::~Tab() {

	disconnect();	
	
	Fluxbox::instance()->removeTabSearch(m_tabwin);
	XDestroyWindow(m_display, m_tabwin);
}


//----------------  createTabWindow ---------------
// (private)
// Creates the Window for tab to be above the title window.
// This should only be called by the constructor.
//-------------------------------------------------
void Tab::createTabWindow() {

	unsigned long attrib_mask = CWBackPixmap | CWBackPixel | CWBorderPixel |
			      CWColormap | CWOverrideRedirect | CWEventMask;
  XSetWindowAttributes attrib;
  attrib.background_pixmap = None;
  attrib.background_pixel = attrib.border_pixel =
			   m_win->screen->getWindowStyle()->tab.border_color.getPixel();
  attrib.colormap = m_win->screen->getColormap();
  attrib.override_redirect = True;
  attrib.event_mask = ButtonPressMask | ButtonReleaseMask |
                      ButtonMotionMask | ExposureMask | EnterWindowMask;
	m_tabwin = XCreateWindow(m_display, m_win->screen->getRootWindow(), 
										m_win->frame.x, m_win->frame.y - m_size_h, 
										m_size_w, m_size_h,
									m_win->screen->getWindowStyle()->tab.border_width, m_win->screen->getDepth(),
                  InputOutput, m_win->screen->getVisual(), attrib_mask, &attrib);
	//set grab
	XGrabButton(m_display, Button1, Mod1Mask, m_tabwin, True,
				ButtonReleaseMask | ButtonMotionMask, GrabModeAsync,
				GrabModeAsync, None, Fluxbox::instance()->getMoveCursor());
	
	
	//save to tabsearch
	Fluxbox::instance()->saveTabSearch(m_tabwin, this);	
	XMoveResizeWindow(m_display, m_tabwin,
									 m_win->frame.x, m_win->frame.y - m_size_h - m_win->screen->getWindowStyle()->tab.border_width_2x,
									 m_size_w, m_size_h);	
	XMapSubwindows(m_display, m_tabwin);
		
	XMapWindow(m_display, m_tabwin);
	
	decorate();
	
}

//-------------- focus --------------------
// Called when the focus changes in m_win 
// updates pixmap or color and draws the tab
//-----------------------------------------
void Tab::focus() {

	if (m_win->focused) {
		if (m_focus_pm)
			XSetWindowBackgroundPixmap(m_display, m_tabwin, m_focus_pm);
		else
			XSetWindowBackground(m_display, m_tabwin, m_focus_pixel);
	} else {
		if (m_unfocus_pm)
			XSetWindowBackgroundPixmap(m_display, m_tabwin, m_unfocus_pm);
		else
			XSetWindowBackground(m_display, m_tabwin, m_unfocus_pixel);
	}
	XClearWindow(m_display, m_tabwin);	
	draw(false);
}

//-------------- raise --------------------
// Raises the tabs in the tablist
//-----------------------------------------
void Tab::raise() {
	//get first tab
	Tab *first = 0;
	first = getFirst(this);
	//raise tabs
	for (; first!=0; first = first->m_next)
		m_win->screen->raiseWindows(&first->m_tabwin, 1);
}

//-------------- decorate --------------------
// decorates the tab with current theme
//--------------------------------------------
void Tab::decorate() {
	
	BImageControl *image_ctrl = m_win->screen->getImageControl();
	Pixmap tmp = m_focus_pm;
  BTexture *texture = &(m_win->screen->getWindowStyle()->tab.l_focus);
 	if (texture->getTexture() & BImage_ParentRelative ) {
		BTexture *pt = &(m_win->screen->getWindowStyle()->tab.t_focus);
		if (pt->getTexture() == (BImage_Flat | BImage_Solid)) {
  		  m_focus_pm = None;
	    m_focus_pixel = pt->getColor()->getPixel();
  	} else
    	m_focus_pm =
	      image_ctrl->renderImage(m_size_w, m_size_h, pt);
  	if (tmp) image_ctrl->removeImage(tmp);
		
	} else {
		if (texture->getTexture() == (BImage_Flat | BImage_Solid)) {
  		  m_focus_pm = None;
	    m_focus_pixel = texture->getColor()->getPixel();
  	} else
    	m_focus_pm =
	      image_ctrl->renderImage(m_size_w, m_size_h, texture);
  	if (tmp) image_ctrl->removeImage(tmp);
	}
	
  tmp = m_unfocus_pm;
  texture = &(m_win->screen->getWindowStyle()->tab.l_unfocus);
  
	if (texture->getTexture() & BImage_ParentRelative ) {
		
		BTexture *pt = &(m_win->screen->getWindowStyle()->tab.t_unfocus);
		if (pt->getTexture() == (BImage_Flat | BImage_Solid)) {
    	m_unfocus_pm = None;
	   	m_unfocus_pixel = pt->getColor()->getPixel();
  	} else
    	m_unfocus_pm =
	      image_ctrl->renderImage(m_size_w, m_size_h, pt);
  	
		
	} else {
		
		if (texture->getTexture() == (BImage_Flat | BImage_Solid)) {
    	m_unfocus_pm = None;
	   	m_unfocus_pixel = texture->getColor()->getPixel();
  	} else
    	m_unfocus_pm =
	      image_ctrl->renderImage(m_size_w, m_size_h, texture);  			
	}
	
	if (tmp) image_ctrl->removeImage(tmp);
	
	XSetWindowBorderWidth(m_display, m_tabwin, m_win->screen->getWindowStyle()->tab.border_width);
	XSetWindowBorder(m_display, m_tabwin, m_win->screen->getWindowStyle()->tab.border_color.getPixel());
}

//-------------- deiconify -------------------
// Deiconifies the tab
// Used from FluxboxWindow to deiconify the tab when the window is deiconfied
//--------------------------------------------
void Tab::deiconify() {
	XMapWindow(m_display, m_tabwin);
}

//------------- iconify --------------------
// Iconifies the tab.
// Used from FluxboxWindow to hide tab win when window is iconified
// disconnects itself from the list
//------------------------------------------
void Tab::iconify() {
	disconnect();
	withdraw();
}

//------------ withdraw --------------
// Unmaps the tab from display
//------------------------------------
void Tab::withdraw() {
	XUnmapWindow(m_display, m_tabwin);	
}

//------------ stick --------------------
// Set/reset the the sticky on all windows in the list
//---------------------------------------
void Tab::stick() {
 
	//get first tab
	Tab *first = 0;
	first = getFirst(this);
	
	//now do stick for all windows in the list
	for (; first!=0; first = first->m_next) {
		FluxboxWindow *win = first->m_win; //just for convenient
		if (win->stuck) {
			win->blackbox_attrib.flags ^= AttribOmnipresent;
			win->blackbox_attrib.attrib ^= AttribOmnipresent;
			win->stuck = false;
			if (!win->iconic)
      	win->screen->reassociateWindow(win, -1, true);
				    
		} else {
			win->stuck = true;
			win->blackbox_attrib.flags |= AttribOmnipresent;
			win->blackbox_attrib.attrib |= AttribOmnipresent;
		}
		win->setState(win->current_state);
	}	

}

//------------- resize -------------
// Resize the window's in the tablist
//----------------------------------
void Tab::resize() {
	Tab *first = getFirst(this);
	//now move and resize the windows in the list
	for (; first != 0; first = first->m_next) {
		if (first!=this) {
			first->m_win->configure(m_win->frame.x, m_win->frame.y,
								m_win->frame.width, m_win->frame.height);
		}
	}
}

//----------- shade --------------
// Shades the windows in the tablist
//--------------------------------
void Tab::shade() {
	for(Tab *first = getFirst(this); first != 0; first = first->m_next) {
		if (first==this)
			continue;
		if (m_win->isShaded())			 
			first->m_win->shade();				
		else
			first->m_win->shade();
	}
}

//------------ draw -----------------
// Draws the tab 
// if pressed = true then it draws the tab in pressed 
// mode else it draws it in normal mode
// note: the "draw in pressed mode" is not implemented yet
//-----------------------------------
void Tab::draw(bool pressed) {
	
	//Draw title string	

	unsigned int l = m_win->client.title_text_w;
	int dlen=m_win->client.title_len;
	int dx=m_win->frame.bevel_w*2;
	
	if (m_win->client.title_text_w > m_size_w) {
		for (; dlen >= 0; dlen--) {
			if (I18n::instance()->multibyte()) {
				XRectangle ink, logical;
				XmbTextExtents(m_win->screen->getWindowStyle()->tab.fontset, 
											m_win->client.title, dlen,
											&ink, &logical);
				l = logical.width;
			} else
				l = XTextWidth(m_win->screen->getWindowStyle()->tab.font, 
									m_win->client.title, dlen);
				l += (m_win->frame.bevel_w * 4);

			if (l < m_size_w)
				break;
		}
	}
	
	switch (m_win->screen->getWindowStyle()->tab.justify) {
	case BScreen::RightJustify:
		dx += m_size_w - l;
		break;

	case BScreen::CenterJustify:
		dx += (m_size_w - l) / 2;
		break;
	}

	//Draw title to m_tabwin

	XClearWindow(m_display, m_tabwin);		
	
	if (I18n::instance()->multibyte()) {
		XmbDrawString(m_display, m_tabwin,
			m_win->screen->getWindowStyle()->tab.fontset,
			((m_win->focused) ? m_win->screen->getWindowStyle()->tab.l_text_focus_gc :
		   m_win->screen->getWindowStyle()->tab.l_text_unfocus_gc), dx, 
			 1 - m_win->screen->getWindowStyle()->tab.fontset_extents->max_ink_extent.y,
			m_win->client.title, dlen);
	} else {
		XDrawString(m_display, m_tabwin,
			((m_win->focused) ? m_win->screen->getWindowStyle()->tab.l_text_focus_gc :
			 m_win->screen->getWindowStyle()->tab.l_text_unfocus_gc), dx,
			m_win->screen->getWindowStyle()->tab.font->ascent + 1, 
			m_win->client.title, dlen);
	}
}

//------------- setPosition -----------------
// Position tab ( follow the m_win pos ).
// Set new position of the other tabs in the chain
//-------------------------------------------
void Tab::setPosition() {	

	//don't do anything if the tablist is freezed
	if (m_stoptabs)
		return;
		
	int pos_x = 0;
	Tab *first = 0;
	int tabs = 0;
	
	m_stoptabs = true; //freeze tablist
	
	//increase pos_x for every tab from start of tablist
	//and check for max tabs
	for (pos_x = 0, first = getFirst(this); 
			first!=0; 
			first = first->m_next, pos_x += m_size_w + m_win->screen->getWindowStyle()->tab.border_width_2x, 
			tabs++) {
			
		if (tabs>=m_max_tabs) {
			cerr<<"Warning: tabs >= max tabs"<<endl;
			break;
		}
		
		//move tab		
		
		XMoveWindow(m_display, first->m_tabwin, 
				m_win->frame.x + pos_x, m_win->frame.y - m_size_h - m_win->screen->getWindowStyle()->tab.border_width_2x);
		
		//dont move fluxboxwindow if the itterator = this
		if (first!=this) {
			first->m_win->configure(m_win->frame.x, m_win->frame.y, 
					m_win->frame.width, m_win->frame.height);
		}	
	}	
	
	m_stoptabs = false;

}

//------------- buttonPressEvent -----------
// Handle button press event here.
//------------------------------------------
void Tab::buttonPressEvent(XButtonEvent *be) {	
	//draw in pressed mode
	draw(true);
	
	//set window to titlewindow so we can take advatage of drag function
	be->window = m_win->frame.title;
	
	//call windows buttonpress eventhandler
	m_win->buttonPressEvent(be);
}

//----------- buttonReleaseEvent ----------
// Handle button release event here.
// If tab is dropped then it should try to find
// the window where the tab where dropped.
//-----------------------------------------
void Tab::buttonReleaseEvent(XButtonEvent *be) {		
	
	if (m_moving ) {
		
		m_moving = false; 
		
		//erase tabmoving rectangle
		XDrawRectangle(m_display, m_win->screen->getRootWindow(), m_win->screen->getOpGC(),
		     m_move_x, m_move_y, 
				 m_size_w, m_size_h);
		
		Fluxbox::instance()->ungrab();
		XUngrabPointer(m_display, CurrentTime);
		
		//storage of window and pos of window where we dropped the tab
		Window child;
		int dest_x = 0, dest_y = 0;
		
		//find window on coordinates of buttonReleaseEvent
		if (XTranslateCoordinates(m_display, m_win->screen->getRootWindow(), 
							m_win->screen->getRootWindow(),
							be->x_root, be->y_root, &dest_x, &dest_y, &child)) {
			
			Tab *tab = 0;
			//search tablist for a tabwindow
			if ((tab = Fluxbox::instance()->searchTab(child))!=0)  {
				//do only attach a hole chain if we dropped the first tab in the dropped chain...
				if (m_prev)
					disconnect();
					
				//attach this tabwindow chain to the tabwindow chain we found.								
				tab->insert(this);
			}	else {				
				disconnect();				
				m_win->configure(be->x_root, be->y_root, m_win->frame.width, m_win->frame.height);
			}
		}
	} else {
		
		//raise this tabwindow
		raise();
		
		//set window to title window soo we can use m_win handler for menu
		be->window = m_win->frame.title;
		
		//call windows buttonrelease event handler so it can popup a menu if needed
		m_win->buttonReleaseEvent(be);
	}
	
}

//------------- exposeEvent ------------
// Handle expose event here.
// Draws the tab unpressed
//--------------------------------------
void Tab::exposeEvent(XExposeEvent *ee) {
	draw(false);
}

//----------- motionNotifyEvent --------
// Handles motion event here
// Draws the rectangle of moving tab
//--------------------------------------
void Tab::motionNotifyEvent(XMotionEvent *me) {
	
	Fluxbox *fluxbox = Fluxbox::instance();
	
	//if mousebutton 2 is pressed
	if (me->state & Button2Mask) {
		if (!m_moving) {
			m_moving = true; 
			
			XGrabPointer(m_display, me->window, False, Button2MotionMask |
								ButtonReleaseMask, GrabModeAsync, GrabModeAsync,
								None, fluxbox->getMoveCursor(), CurrentTime);

			fluxbox->grab();
			
			m_move_x = m_win->frame.x;
			m_move_y = m_win->frame.y;
				
			XDrawRectangle(m_display, m_win->screen->getRootWindow(), 
					m_win->screen->getOpGC(),
					m_move_x, m_move_y,
					m_size_w, m_size_h);
	
		} else {
		
			int dx = me->x_root - 1, dy = me->y_root - 1;

			dx -= m_win->screen->getBorderWidth();
			dy -= m_win->screen->getBorderWidth();

			if (m_win->screen->getEdgeSnapThreshold()) {
				int drx = m_win->screen->getWidth() - (dx + 1);

				if (dx > 0 && dx < drx && dx < m_win->screen->getEdgeSnapThreshold()) 
					dx = 0;
				else if (drx > 0 && drx < m_win->screen->getEdgeSnapThreshold())
					dx = m_win->screen->getWidth() - 1;

				int dtty, dbby, dty, dby;
		
				switch (m_win->screen->getToolbarPlacement()) {
				case Toolbar::TopLeft:
				case Toolbar::TopCenter:
				case Toolbar::TopRight:
					dtty = m_win->screen->getToolbar()->getExposedHeight() +
							m_win->screen->getBorderWidth();
					dbby = m_win->screen->getHeight();
					break;

				default:
					dtty = 0;
					dbby = m_win->screen->getToolbar()->getY();
					break;
				}
		
				dty = dy - dtty;
				dby = dbby - (dy + 1);

				if (dy > 0 && dty < m_win->screen->getEdgeSnapThreshold())
					dy = dtty;
				else if (dby > 0 && dby < m_win->screen->getEdgeSnapThreshold())
					dy = dbby - 1;
		
			}
		
			//erase rectangle
			XDrawRectangle(m_display, m_win->screen->getRootWindow(),
					m_win->screen->getOpGC(),
					m_move_x, m_move_y, 
					m_size_w, m_size_h);

			//redraw rectangle at new pos
			m_move_x = dx;
			m_move_y = dy;			
			XDrawRectangle(m_display, m_win->screen->getRootWindow(), 
					m_win->screen->getOpGC(),
					m_move_x, m_move_y,
					m_size_w, m_size_h);					
			
		}
	} 
}

//-------------- getFirst() ---------
// Returns the first Tab in the chain 
// of currentchain. 
//-----------------------------------
Tab *Tab::getFirst(Tab *current) {
	for (; current->m_prev != 0; current = current->m_prev);
	return current;
}

//-------------- insert ------------
// (private)
// Inserts a tab in the chain
//----------------------------------
void Tab::insert(Tab *tab) {
	
	if (!tab || tab == this) //dont insert if the tab = 0 or the tab = this
		return;

	Tab *first = getFirst(this);
	
	//if the tab already in chain then disconnect it
	for (; first!=0; first = first->m_next) {
		if (first==tab) {
			#ifdef DEBUG
			cerr<<"Tab already in chain. Disconnecting!"<<endl;			
			#endif
			tab->disconnect();
			break;
		}
	}

	//get last tab in the chain to be inserted
	Tab *last = tab;
	for (; last->m_next!=0; last=last->m_next); 
	//do sticky before we connect it to the chain
	//sticky bit on window
	if (m_win->isStuck() && !tab->m_win->isStuck() ||
			!m_win->isStuck() && tab->m_win->isStuck())
			tab->m_win->stick(); //this will set all the m_wins in the list
	
	//connect the tab to this chain
	
	if (m_next)	
		m_next->m_prev = last;
	tab->m_prev = this; 
	last->m_next = m_next; 
	
	m_next = tab;	
	
	//move and resize all windows in the tablist we inserted
	//only from first tab of the inserted chain to the last
	for (; tab!=last->m_next; tab=tab->m_next) {
		tab->m_win->configure(m_win->frame.x, m_win->frame.y,
								m_win->frame.width, m_win->frame.height);	
		
		if (m_win->isShaded() && !tab->m_win->isShaded()||
				!m_win->isShaded() && tab->m_win->isShaded())
			tab->m_win->shade();
		
	}
	
}

//---------- disconnect() --------------
// Disconnects the tab from any chain
//--------------------------------------
void Tab::disconnect() {
	Tab *tmp = 0;
	if (m_prev) {	//if this have a chain to "the left" (previous tab) then set it's next to this next
		m_prev->m_next = m_next;
		tmp = m_prev;
	} 
	if (m_next) {	//if this have a chain to "the right" (next tab) then set it's prev to this prev
		m_next->m_prev = m_prev;
		tmp = m_next;
	}
	
		
	//mark as no chain, previous and next.
	m_prev = 0;
	m_next = 0;
	
	//reposition the tabs
	if (tmp)
		tmp->setPosition();

	setPosition();
}
